/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SVO_DEBUG			FALSE
#define FE_SVO_THREAD_DEBUG		FALSE
#define FE_SVO_RECORDER			FALSE

#include "oplab/oplab.pmh"

using namespace fe;

SurfaceViewerOp::SurfaceViewerOp(void):
	m_jobs(0),
	m_lastFrame(-1.0)
{
#if FE_SVO_DEBUG
	feLog("SurfaceViewerOp::SurfaceViewerOp\n");
#endif
}

SurfaceViewerOp::~SurfaceViewerOp(void)
{
#if FE_SVO_DEBUG
	feLog("SurfaceViewerOp::~SurfaceViewerOp\n");
#endif
}

void SurfaceViewerOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<SpatialVector>("vertical")=SpatialVector(0.0,1.0,0.0);

	catalog<String>("drawStyle")="solid";
	catalog<String>("drawStyle","choice:0")="pointCloud";
	catalog<String>("drawStyle","label:0")="Point Cloud";
	catalog<String>("drawStyle","choice:1")="wireframe";
	catalog<String>("drawStyle","label:1")="Wireframe";
	catalog<String>("drawStyle","choice:2")="outline";
	catalog<String>("drawStyle","label:2")="Outline";
	catalog<String>("drawStyle","choice:3")="solid";
	catalog<String>("drawStyle","label:3")="Solid";
	catalog<String>("drawStyle","choice:4")="edgedSolid";
	catalog<String>("drawStyle","label:4")="Edged Solid";
	catalog<String>("drawStyle","choice:5")="foreshadow";
	catalog<String>("drawStyle","label:5")="Foreshadow";

	catalog<Real>("lineWidth")=1.0;
	catalog<Real>("lineWidth","high")=8.0;
	catalog<String>("lineWidth","label")="Line Width";

	catalog<bool>("finished")=FALSE;
	catalog<String>("finished","IO")="output";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Input Surface 2");
	catalog<bool>("Input Surface 2","optional")=TRUE;

	catalog< sp<Component> >("Input Surface 3");
	catalog<bool>("Input Surface 3","optional")=TRUE;

	catalog< sp<Component> >("Input Surface 4");
	catalog<bool>("Input Surface 4","optional")=TRUE;

	//* no auto-copy
	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<bool>("Output Surface","recycle")=true;
}

void SurfaceViewerOp::post(Real a_frame)
{
#if	FE_SVO_THREAD_DEBUG
	feLog("SurfaceViewerOp::post %.6G jobs %d\n",a_frame,m_jobs);
#endif

	m_spGang->post(a_frame);
	m_jobs++;

	Real completed=Real(0);
	while(m_jobs)
	{
		while(m_spGang->acceptDelivery(completed))
		{
#if	FE_SVO_THREAD_DEBUG
			feLog("SurfaceViewerOp::post accept %.6G jobs %d\n",
					completed,m_jobs);
#endif

			m_jobs--;
		}
		milliSleep(1);
	}

#if	FE_SVO_THREAD_DEBUG
	feLog("SurfaceViewerOp::post done jobs %d\n",m_jobs);
#endif
}

void SurfaceViewerOp::Worker::operate(void)
{
#if	FE_SVO_THREAD_DEBUG
	feLog("SurfaceViewerOp::Worker::operate\n");
#endif

	BWORD bound=FALSE;
	BWORD firstFrame=TRUE;

	Real job=Real(0);
	while(TRUE)
	{
		//* lock down for each iteration
		sp<SurfaceViewerOp> spSurfaceViewerOp=m_hpSurfaceViewerOp;
		if(spSurfaceViewerOp.isNull())
		{
			break;
		}

		sp<QuickViewerI> spQuickViewer=spSurfaceViewerOp->quickViewer();
		if(spQuickViewer.isNull())
		{
			milliSleep(10);
			continue;
		}

		if(!bound)
		{
			spQuickViewer->open();

			sp<DrawI> spDrawI=spQuickViewer->getDrawI();
			if(spDrawI.isNull())
			{
				continue;
			}

			spSurfaceViewerOp->drawHandler()->bind(spDrawI);

			sp<DrawMode> spDrawMode=spDrawI->drawMode();
			if(spDrawMode.isValid())
			{
				spDrawMode->setBackfaceCulling(FALSE);
				spDrawMode->setFrontfaceCulling(FALSE);
			}

			sp<HandlerI> spHandlerI(spDrawI);
			if(spHandlerI.isValid())
			{
				spQuickViewer->insertEventHandler(spDrawI);
			}

			bound=TRUE;
		}

		sp<DrawI> spDrawI=spQuickViewer->getDrawI();
		if(spDrawI.isValid())
		{
			sp<DrawMode> spDrawMode=spDrawI->drawMode();
			if(spDrawMode.isValid())
			{
				spDrawMode->setDrawStyle(spSurfaceViewerOp->drawStyle());
				spDrawMode->setLineWidth(spSurfaceViewerOp->lineWidth());
			}
		}


#if	FE_SVO_THREAD_DEBUG
		feLog("SurfaceViewerOp::Worker::operate check\n");
#endif

#if FE_SVO_RECORDER
		m_hpJobQueue->waitForJob(job);
#else
		if(m_hpJobQueue->take(job))
#endif
		{
			if(job<0)
			{
#if	FE_SVO_THREAD_DEBUG
				feLog("SurfaceViewerOp::Worker::operate quit %.6G\n",job);
#endif

				m_hpJobQueue->deliver(job);
				break;
			}

#if	FE_SVO_THREAD_DEBUG
			feLog("SurfaceViewerOp::Worker::operate do job %.6G\n",job);
#endif

			const U32 us=systemMicroseconds();
			const I32 delta=us-m_us;
			m_us=us;

#if	FE_SVO_THREAD_DEBUG
			feLog("SurfaceViewerOp::Worker::operate delta %d us\n",delta);
#endif

			sp<DrawableI> spDrawable1=spSurfaceViewerOp->drawable1();
			sp<DrawableI> spDrawable2=spSurfaceViewerOp->drawable2();
			sp<DrawableI> spDrawable3=spSurfaceViewerOp->drawable3();
			sp<DrawableI> spDrawable4=spSurfaceViewerOp->drawable4();

			spSurfaceViewerOp->drawHandler()->setFrame(job);
			spSurfaceViewerOp->drawHandler()->setMicroseconds(delta);
			spSurfaceViewerOp->drawHandler()->bindDrawable1(spDrawable1);
			spSurfaceViewerOp->drawHandler()->bindDrawable2(spDrawable2);
			spSurfaceViewerOp->drawHandler()->bindDrawable3(spDrawable3);
			spSurfaceViewerOp->drawHandler()->bindDrawable4(spDrawable4);

			sp<SurfaceI> spSurface=spDrawable1;
			if(spSurface.isValid())
			{
				const SpatialVector center=spSurface->center();
				const Real radius=spSurface->radius();

				const Real minDistance=
						(firstFrame? 10.0: 2.0)*radius;			//* NOTE tweak
				const Real maxOffset=
						(firstFrame? 0.0: 0.3)*radius;			//* NOTE tweak
				const Real response=firstFrame? 1.0: 0.02;		//* NOTE tweak

				sp<ControllerI> spController=
						spQuickViewer->getControllerI();
				SpatialVector lookPoint=spController->lookPoint();
				const SpatialVector correction=center-lookPoint;
				const Real offset=magnitude(correction);
				const Real scalar=offset-maxOffset;
				if(scalar>0.0)
				{
					lookPoint+=correction*(response*scalar/offset);
					spController->setLookPoint(lookPoint);
				}

				Real distance=spController->distance();
				if(distance<minDistance)
				{
					distance=(1.0-response)*distance+response*minDistance;

					const Real azimuth=spController->azimuth();
					const Real downangle=spController->downangle();

					spController->setLookAngle(azimuth,downangle,distance);
				}

//				feLog("look at %s dist %.6G\n",c_print(lookPoint),distance);

				firstFrame=FALSE;
			}

#if	FE_SVO_THREAD_DEBUG
			feLog("SurfaceViewerOp::Worker::operate done job %.6G\n",job);
#endif

			m_hpJobQueue->deliver(job);

#if	FE_SVO_THREAD_DEBUG
			feLog("SurfaceViewerOp::Worker::operate delivered job %.6G\n",job);
#endif
		}
#if	FE_SVO_THREAD_DEBUG
		else
		{
			feLog("SurfaceViewerOp::Worker::operate no jobs\n");
		}
#endif

		sp<QuickViewerI> spQuickViewerI=spSurfaceViewerOp->quickViewer();
		if(spQuickViewerI.isValid())
		{
			spQuickViewerI->run(1);
		}
	}

#if	FE_SVO_THREAD_DEBUG
	feLog("SurfaceViewerOp::Worker::operate done\n");
#endif
}

void SurfaceViewerOp::handle(Record& a_rSignal)
{
	const String drawStyle=catalog<String>("drawStyle");

	m_drawStyle=DrawMode::e_solid;
	if(drawStyle=="pointCloud")
	{
		m_drawStyle=DrawMode::e_pointCloud;
	}
	else if(drawStyle=="wireframe")
	{
		m_drawStyle=DrawMode::e_wireframe;
	}
	else if(drawStyle=="outline")
	{
		m_drawStyle=DrawMode::e_outline;
	}
	else if(drawStyle=="edgedSolid")
	{
		m_drawStyle=DrawMode::e_edgedSolid;
	}
	else if(drawStyle=="foreshadow")
	{
		m_drawStyle=DrawMode::e_foreshadow;
	}

	m_lineWidth=catalog<Real>("lineWidth");

	SpatialVector vertical=unitSafe(catalog<SpatialVector>("vertical"));
	if(magnitude(vertical)<0.999)
	{
		set(vertical,0.0,1.0,0.0);
	}

#if FE_SVO_DEBUG
	feLog("SurfaceViewerOp::handle\n");
#endif

	if(m_spGang.isNull())
	{
		m_spGang=new Gang<Worker,Real>();
		m_spGang->start(sp<Counted>(this),1);
	}

	//* see surface/test/xSurface.cc

	const Real frame=currentFrame(a_rSignal);
	if(m_lastFrame!=frame)
	{
		m_spDrawable1=NULL;
		m_spDrawable2=NULL;
		m_spDrawable3=NULL;
		m_spDrawable4=NULL;
		m_lastFrame=frame;
	}

	if(m_spDrawable1.isNull())
	{
		sp<SurfaceI> spInput;
		if(!access(spInput,"Input Surface")) return;

		m_spDrawable1=spInput;
	}
	if(m_spDrawable2.isNull())
	{
		sp<SurfaceI> spInput;
		access(spInput,"Input Surface 2",e_quiet);

		m_spDrawable2=spInput;
	}
	if(m_spDrawable3.isNull())
	{
		sp<SurfaceI> spInput;
		access(spInput,"Input Surface 3",e_quiet);

		m_spDrawable3=spInput;
	}
	if(m_spDrawable4.isNull())
	{
		sp<SurfaceI> spInput;
		access(spInput,"Input Surface 4",e_quiet);

		m_spDrawable4=spInput;
	}

	if(m_spDrawable1.isNull())
	{
		catalog<String>("error")="input surface not drawable";
		return;
	}

	if(m_spDrawHandler.isNull())
	{
		m_spDrawHandler=
				Library::create<DrawHandler>("SurfaceViewerOp::DrawHandler");
	}

#if FE_SVO_RECORDER
	//* TODO param
	if(m_spVideoRecorder.isNull())
	{
		m_spVideoRecorder=registry()->create("*.VideoRecorder");
		m_spVideoRecorder->configure("test/chainvideo",240);
	}
#endif

	if(m_spQuickViewerI.isNull())
	{
		registry()->master()->catalog()->catalog<String>("hint:DrawI")=
				"*.DrawCached";

		m_spQuickViewerI=registry()->create("QuickViewerI");
		if(!m_spQuickViewerI.isValid())
		{
			catalog<String>("error")="couldn't create viewer";
			return;
		}

		sp<ControllerI> spController=m_spQuickViewerI->getControllerI();
		spController->setVertical(vertical);
		spController->setLookPoint(SpatialVector(0.0,0.0,0.0));
		spController->setLookAngle(-80.0,30.0,1.0f);

		m_spQuickViewerI->insertDrawHandler(m_spDrawHandler);

		if(m_spVideoRecorder.isValid())
		{
			feLog("SurfaceViewerOp::handle adding VideoRecorder\n");

			m_spQuickViewerI->insertDrawHandler(m_spVideoRecorder);
		}
	}

	BWORD closed=FALSE;

	if(m_spQuickViewerI.isValid())
	{
		sp<ControllerI> spController=m_spQuickViewerI->getControllerI();
		closed=spController->closeEvent();
	}

	post(closed? -1: frame);

	catalog<bool>("finished")=closed;

#if FE_SVO_DEBUG
	feLog("SurfaceViewerOp::handle done\n");
#endif
}

void SurfaceViewerOp::DrawHandler::handle(Record& a_render)
{
	if(!m_asViewer.scope().isValid())
	{
		m_asViewer.bind(a_render.layout()->scope());
	}
	if(!m_asViewer.viewer_perspective.check(a_render) ||
			!m_asViewer.viewer_layer.check(a_render) ||
			!m_asViewer.viewer_layer(a_render))
	{
		return;
	}

	const I32& rLayer=m_asViewer.viewer_layer(a_render);
	if(rLayer==1)
	{
		//* main render

		if(m_spDrawable1.isValid())
		{
			m_spDrawable1->draw(m_spDrawI,NULL);
		}

		if(m_spDrawable2.isValid())
		{
			m_spDrawable2->draw(m_spDrawI,NULL);
		}

		if(m_spDrawable3.isValid())
		{
			m_spDrawable3->draw(m_spDrawI,NULL);
		}

		if(m_spDrawable4.isValid())
		{
			m_spDrawable4->draw(m_spDrawI,NULL);
		}

		return;
	}

	if(rLayer==2)
	{
		const Color yellow(1.0,1.0,0.0);
		const SpatialVector location1(10.0,30.0,0.0);
		const SpatialVector location2(10.0,20.0,0.0);

		String text;

		if(m_us>0)
		{
			text.sPrintf("%.1f ms",0.001*m_us);
			m_spDrawI->drawAlignedText(location1,text,yellow);
		}

		text.sPrintf("Frame %.6G",m_frame);
		m_spDrawI->drawAlignedText(location2,text,yellow);

		return;
	}
}
