/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SVDB_DEBUG			FALSE
#define FE_SVDB_IMPACT_DEBUG	FALSE

#include "vdb.pmh"

namespace fe
{

SurfaceVDB::SurfaceVDB(void)
{
#if	FE_SVDB_DEBUG
	feLog("SurfaceVDB::SurfaceVDB\n");
#endif
}

SurfaceVDB::~SurfaceVDB(void)
{
#if	FE_SVDB_DEBUG
	feLog("SurfaceVDB::~SurfaceVDB\n");
#endif
}

Protectable* SurfaceVDB::clone(Protectable* pInstance)
{
	return NULL;
}

void SurfaceVDB::bind(Instance a_instance)
{
	void* pPtr=a_instance.cast<void*>();
	openvdb::GridBase::ConstPtr* pGridPtr=(openvdb::GridBase::ConstPtr*)pPtr;

	bindGridBase(*pGridPtr);
}

//* TODO don't generate CPT until needed
void SurfaceVDB::bindGridBase(openvdb::GridBase::ConstPtr& a_gridPtr)
{
	const openvdb::Vec3R origin(0.0,0.0,0.0);

#if	FE_SVDB_DEBUG
	feLog("\nSurfaceVDB::bindGridBase \"%s\" type \"%s\" valueType \"%s\"\n",
			a_gridPtr->getName().c_str(),
			a_gridPtr->type().c_str(),
			a_gridPtr->valueType().c_str());
#endif

	if(a_gridPtr->getName()=="normal")
	{
		m_normalGridPtr=
				openvdb::gridConstPtrCast<openvdb::Vec3SGrid>(a_gridPtr);
	}
	else if(a_gridPtr->getName()=="uv")
	{
		m_uvGridPtr=
				openvdb::gridConstPtrCast<openvdb::Vec3SGrid>(a_gridPtr);
	}
	else if(a_gridPtr->getName()=="surface")
	{
		m_sdfGridPtr=
				openvdb::gridConstPtrCast<openvdb::FloatGrid>(a_gridPtr);

		openvdb::CoordBBox bbox=m_sdfGridPtr->evalActiveVoxelBoundingBox();
		openvdb::Vec3d v3dCenter=bbox.getCenter();
		openvdb::Vec3d min=bbox.min().asVec3d();
		openvdb::Vec3d max=bbox.max().asVec3d();
		openvdb::Vec3d voxelSize=m_sdfGridPtr->voxelSize();

		set(m_center,v3dCenter[0]*voxelSize[0],
				v3dCenter[1]*voxelSize[1],v3dCenter[2]*voxelSize[2]);
		const SpatialVector corner(max[0]*voxelSize[0],
				max[1]*voxelSize[1],max[2]*voxelSize[2]);
		m_radius=magnitude(corner-m_center);

		openvdb::tools::Cpt<openvdb::FloatGrid> cpt(*m_sdfGridPtr);
		m_cptGridPtr=cpt.process(true,false);

#if	FE_SVDB_DEBUG
		openvdb::tools::BoxSampler sampler;

		const Real floatValue=sampler.sample(m_sdfGridPtr->tree(),origin);

		openvdb::Vec3s v3sValue;
		sampler.sample(m_cptGridPtr->tree(),origin,v3sValue);

		feLog("SurfaceVDB::bindGridBase at origin,"
				" value %.6G cpt %.6G %.6G %.6G\n",
				floatValue,c_print(m_center),
				v3sValue[0],v3sValue[1],v3sValue[2]);

		bool uniform=m_sdfGridPtr->hasUniformVoxels();

		feLog("  center %s size %.6G %.6G %.6G u %d\n",c_print(m_center),
				voxelSize[0],voxelSize[1],voxelSize[2],uniform);
		feLog("  min %.6G %.6G %.6G max %.6G %.6G %.6G\n",
				min[0],min[1],min[2],max[0],max[1],max[2]);
#endif
	}
}

SurfaceI::Containment SurfaceVDB::containment(
	const SpatialVector& a_origin) const
{
	openvdb::Vec3d voxelSize=m_cptGridPtr->voxelSize();

	const openvdb::Vec3R v3rOrigin(a_origin[0]/voxelSize[0],
			a_origin[1]/voxelSize[1],a_origin[2]/voxelSize[2]);

	openvdb::tools::PointSampler sampler;

	openvdb::Vec3s cptValue;
	const BWORD successCPT=sampler.sample(m_cptGridPtr->tree(),
			v3rOrigin,cptValue);

	//* NOTE exact zero
	if(cptValue[0]==0.0 && cptValue[1]==0.0 && cptValue[2]==0.0)
	{
		return SurfaceI::e_unknown;
	}

	Real sdfSample=0.0;
	const BWORD successSDF=sampler.sample(m_sdfGridPtr->tree(),
			v3rOrigin,sdfSample);

	return sdfSample>0.0? SurfaceI::e_outside: SurfaceI::e_inside;
}

sp<SurfaceI::ImpactI> SurfaceVDB::nearestPoint(
	const SpatialVector& a_origin) const
{
	openvdb::Vec3d voxelSize=m_cptGridPtr->voxelSize();

	const openvdb::Vec3R v3rOrigin(a_origin[0]/voxelSize[0],
			a_origin[1]/voxelSize[1],a_origin[2]/voxelSize[2]);

	//* NOTE BoxSampler and QuadraticSampler can give lousy results
//	openvdb::tools::BoxSampler sampler;
//	openvdb::tools::QuadraticSampler sampler;
	openvdb::tools::PointSampler sampler;

	Real sdfSample=0.0;
	const BWORD successSDF=sampler.sample(m_sdfGridPtr->tree(),
			v3rOrigin,sdfSample);

	openvdb::Vec3s cptValue;
	const BWORD successCPT=sampler.sample(m_cptGridPtr->tree(),
			v3rOrigin,cptValue);

	//* NOTE exact zero
	if(cptValue[0]==0.0 && cptValue[1]==0.0 && cptValue[2]==0.0)
	{
		return sp<ImpactI>(NULL);
	}

	const SpatialVector cptSample(cptValue[0]*voxelSize[0],
			cptValue[1]*voxelSize[1],cptValue[2]*voxelSize[2]);

/*
	openvdb::Vec3s normalValue;
	const BWORD successNormal=sampler.sample(m_normalGridPtr->tree(),
			v3rOrigin,normalValue);
*/

	const SpatialVector towards=unitSafe(cptSample-a_origin);
	const SpatialVector intersection=a_origin+fabs(sdfSample)*towards;

	//* NOTE default to lousy normal using spherical approximation
	SpatialVector normal=unitSafe(intersection-m_center);
	if(m_normalGridPtr!=NULL)
	{
		openvdb::Vec3s normalValue;
		const BWORD successNormal=sampler.sample(m_normalGridPtr->tree(),
				v3rOrigin,normalValue);
		set(normal,normalValue[0],normalValue[1],normalValue[2]);
	}

	SpatialVector uv(0.0,0.0,0.0);
	if(m_uvGridPtr!=NULL)
	{
		openvdb::Vec3s uvValue;
		const BWORD successUV=sampler.sample(m_uvGridPtr->tree(),
				v3rOrigin,uvValue);
		set(uv,uvValue[0],uvValue[1],uvValue[2]);
	}

	SpatialVector direction=intersection-a_origin;
	const Real distance=magnitude(direction);
	if(distance>0.0)
	{
		direction*=1.0/distance;
	}

#if	FE_SVDB_IMPACT_DEBUG
	feLog("value %.6G %.6G %.6G success %d %d size %.6G %.6G %.6G\n",
			cptValue[0],cptValue[1],cptValue[2],successSDF,successCPT,
			voxelSize[0],voxelSize[1],voxelSize[2]);
	feLog("near %s sdf %.6G cpt %s\n",c_print(a_origin),
			sdfSample,c_print(cptSample));
	feLog("intersect %s dist %.6G dir %s\n",c_print(intersection),
			distance,c_print(direction));
	feLog("normal %s uv %s\n",c_print(normal),c_print(uv));
#endif

	sp<SurfaceVDB::Impact> spImpact=m_vdbImpactPool.get();
	FEASSERT(spImpact.isValid());

	spImpact->setSurface(this);
	spImpact->setLocationLocal(m_center);
	spImpact->setIntersectionLocal(intersection);
	spImpact->setNormalLocal(normal);
	spImpact->setUV(uv);
	spImpact->setDiffuse(m_diffuse);
	spImpact->setRadius(m_radius);
	spImpact->setOrigin(a_origin);
	spImpact->setDirection(direction);
	spImpact->setDistance(distance);

	return spImpact;
}

void SurfaceVDB::draw(const SpatialTransform& a_rTransform,
	sp<DrawI> a_spDrawI,const fe::Color* a_color) const
{
	openvdb::FloatGrid::ConstAccessor floatAccessor=
			m_sdfGridPtr->getConstAccessor();

	openvdb::Vec3SGrid::ConstAccessor* pV3sAccessor=NULL;
	if(m_normalGridPtr!=NULL)
	{
		pV3sAccessor=new openvdb::Vec3SGrid::ConstAccessor(
				m_normalGridPtr->getConstAccessor());
	}

	openvdb::Vec3d voxelSize=m_cptGridPtr->voxelSize();

	for(openvdb::Vec3SGrid::ValueOnCIter it=m_cptGridPtr->cbeginValueOn();
			it.test();++it)
	{
		const openvdb::Vec3s& cptValue=it.getValue();

		if(it.isVoxelValue())
		{
			//* leaf (voxel)
			openvdb::Coord coord=it.getCoord();
			openvdb::Vec3s v3s=coord.asVec3s();

			float signedDistance=0.0;
			const BWORD sdfActive=
					floatAccessor.probeValue(coord,signedDistance);

			if(!sdfActive)
			{
				continue;
			}

			SpatialVector normal(0.0,1.0,0.0);
			if(pV3sAccessor)
			{
				openvdb::Vec3s normalValue;
				const BWORD normalActive=
						pV3sAccessor->probeValue(coord,normalValue);
				set(normal,normalValue[0],normalValue[1],normalValue[2]);
			}

//			const openvdb::tree::LeafNode<float,3>* pLeaf=
//					floatAccessor.probeConstLeaf(coord);
//			const openvdb::tree::LeafNode<openvdb::Vec3s,3>* pLeaf=
//					v3sAccessor.probeConstLeaf(coord);

			const SpatialVector voxelCenter(v3s[0]*voxelSize[0],
					v3s[1]*voxelSize[1],v3s[2]*voxelSize[2]);
			const SpatialVector target(cptValue[0]*voxelSize[0],
					cptValue[1]*voxelSize[1],cptValue[2]*voxelSize[2]);
			const SpatialVector towards=unitSafe(target-voxelCenter);
			const SpatialVector retarget=
					voxelCenter+fabs(signedDistance)*towards;

#if FALSE
			if(voxelCenter[1]<3.0)
			{
				continue;
			}
			if(magnitude(retarget)>magnitude(voxelCenter))
			{
				feLog(" leaf %s ret %s sdf %.6G cpt %.6G %.6G %.6G\n",
						c_print(voxelCenter),c_print(retarget),
						signedDistance,cptValue[0],cptValue[1],cptValue[2]);
				feLog(" size %d values %d on %d onLeaf %d empty %d dense %d\n",
						pLeaf->size(),
						pLeaf->numValues(),
						pLeaf->onVoxelCount(),
						pLeaf->onLeafVoxelCount(),
						pLeaf->isEmpty(),
						pLeaf->isDense());
			}
#endif

			SpatialVector line[2];
			line[0]=voxelCenter;
			line[1]=retarget;

			const Color black(0.0,0.0,0.0);
			const Color white(1.0,1.0,1.0);
			const Color blue(0.0,0.0,1.0);
			const Color green(0.0,1.0,0.0);
			const Color red(1.0,0.0,0.0);
			Color colors[2];
			colors[0]=white;
			colors[1]=(signedDistance>0.0)? green: blue;

			a_spDrawI->drawLines(line,NULL,2,
					DrawI::e_discrete,TRUE,colors);

			line[1]=line[0]+normal;

			colors[0]=black;
			colors[1]=red;

			a_spDrawI->drawLines(line,NULL,2,
					DrawI::e_discrete,TRUE,colors);
		}

#if 0
		openvdb::CoordBBox bbox;
		it.getBoundingBox(bbox);
		openvdb::Vec3d v3d=bbox.getCenter();
		feLog(" bbox %.6G %.6G %.6G\n",v3d[0],v3d[1],v3d[2]);
#endif
	}

	delete pV3sAccessor;
}

} // namespace
